"""Init file for the _isomorphism submodule project."""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# This file is a modification and extension of the [GNU LPGL] licensed
# PyBliss which can be found at: http://www.tcs.hut.fi/Software/bliss/
# PyBliss and Bliss are copyright of their respective owners.
# License: BSD 3 clause"
from grakel.kernels._isomorphism.bliss import Graph

__all__ = [
    "Graph",
]
